/*
 *
 * SdbVariable.c --
 *
 *      This file contains functions that support the organization
 *      and manipulation of database variables.
 *
 * Copyright (c) 1995 Tom Moore.  All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice appear in all copies of this software.
 * 
 */

#include "sdbInt.h"
#include <unistd.h>
#include <fcntl.h>


/*
 *------------------------------------------------------------
 *
 * SdbFound --
 *
 *      Search the variable dictionary for a name
 *
 * Results:
 *      Returns the address of the variable, or NULL if not found.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------
 */
SdbVariable            *
SdbFound(V, string)
     SdbVariable            *V;
     CONST char             *string;
{
    SdbVariable            *varPtr;

    if (!V)
        return NULL;

    for (varPtr = V; varPtr; varPtr = varPtr->next)
        if (strcmp(varPtr->id, string) == 0)
            return varPtr;
    return NULL;
}

/*
 *------------------------------------------------------------
 *
 * SdbPrettyPrint --
 *
 *      Format an expression from byte codes into a string.
 *
 * Results:
 *      A standard TCL error code
 *
 * Side effects:
 *      Appends the string to the interpretor result.
 *
 *------------------------------------------------------------
 */
int
SdbPrettyPrint(interp, vars, tree)
     Tcl_Interp             *interp;
     SdbVariable            *vars;
     SdbNode                *tree;
{

    if (!tree)
        return TCL_OK;
    switch (tree->type) {
    case booleanOp:
    case binaryOp:
    case stringCompare:
        Tcl_AppendResult(interp, "(", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->left);
        switch (tree->content.op) {
        case opSum:
            Tcl_AppendResult(interp, "+", (char *) NULL);
            break;
        case opSub:
            Tcl_AppendResult(interp, "-", (char *) NULL);
            break;
        case opMul:
            Tcl_AppendResult(interp, "*", (char *) NULL);
            break;
        case opDiv:
            Tcl_AppendResult(interp, "/", (char *) NULL);
            break;
        case opGe:
            Tcl_AppendResult(interp, ">=", (char *) NULL);
            break;
        case opGt:
            Tcl_AppendResult(interp, ">", (char *) NULL);
            break;
        case opLe:
            Tcl_AppendResult(interp, "<=", (char *) NULL);
            break;
        case opLt:
            Tcl_AppendResult(interp, "<", (char *) NULL);
            break;
        case opEq:
            Tcl_AppendResult(interp, "==", (char *) NULL);
            break;
        case opNe:
            Tcl_AppendResult(interp, "!=", (char *) NULL);
            break;
        case opAnd:
            Tcl_AppendResult(interp, "&&", (char *) NULL);
            break;
        case opOr:
            Tcl_AppendResult(interp, "||", (char *) NULL);
            break;
        default:
            Tcl_AppendResult(interp, " error! ", (char *) NULL);
            break;
        }
        SdbPrettyPrint(interp, vars, tree->right);
        Tcl_AppendResult(interp, ")", (char *) NULL);
        break;


    case constValue:{
            char                    temp[50];

            sprintf(temp, "%-.15g", tree->content.value);
            Tcl_AppendResult(interp, temp, (char *) NULL);
            break;
        }

    case constCode:
        Tcl_AppendResult(interp, "\"", tree->content.code, "\"",
                (char *) NULL);
        break;

    case unaryOp:
        switch (tree->content.op) {
        case opNeg:
            Tcl_AppendResult(interp, "-", (char *) NULL);
            break;
        case opSum:
            Tcl_AppendResult(interp, "+", (char *) NULL);
            break;
        case opNot:
            Tcl_AppendResult(interp, "!", (char *) NULL);
            break;
        default:
            Tcl_AppendResult(interp, " error! ", (char *) NULL);
            break;
        }
        SdbPrettyPrint(interp, vars, tree->right);
        break;

    case charFnct:
        Tcl_AppendResult(interp, SdbWhatFnct(tree->content.fnctptr), "(",
                (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        Tcl_AppendResult(interp, ")", NULL);
        break;

    case simplexFnct:
        Tcl_AppendResult(interp, SdbWhatFnct(tree->content.fnctptr), "(",
                (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        Tcl_AppendResult(interp, ")", NULL);
        break;

    case duplexFnct:
        Tcl_AppendResult(interp, SdbWhatFnct(tree->content.fnctptr), "(",
                (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->left);
        Tcl_AppendResult(interp, ",", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        Tcl_AppendResult(interp, ")", (char *) NULL);
        break;

    case hsgFnct:
        Tcl_AppendResult(interp, SdbWhatFnct(tree->content.fnctptr), "(",
                (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        Tcl_AppendResult(interp, ")", (char *) NULL);
        break;

    case charSequence:
    case numSequence:
        if (tree->left) {
            SdbPrettyPrint(interp, vars, tree->left);
            Tcl_AppendResult(interp, ";", (char *) NULL);
        }
        SdbPrettyPrint(interp, vars, tree->right);
        break;

    case assignChar:
    case assignNum:
        SdbPrettyPrint(interp, vars, tree->left);
        Tcl_AppendResult(interp, "=", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        break;

    case formatFnct: {
        SdbNode *next;
        Tcl_AppendResult(interp, "format(", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->left);
	next = tree->right;
	while(next) {
	  Tcl_AppendResult(interp, ",", (char *) NULL);
	  SdbPrettyPrint(interp, vars, next->left);
	  next = next->right;
	}
	Tcl_AppendResult(interp, ")", (char *) NULL);
	break;
      }

    case numExpr:
    case charExpr:
    case hsgCodeIndex:
    case hsgNumberIndex:
        Tcl_AppendElement(interp, (tree->content.varPtr)->id);
        break;

    case hsgCode:
    case hsgNumber:
        Tcl_AppendElement(interp, tree->content.code);
        break;

    case hsgCodeList:{
            int                     i;

            Tcl_AppendResult(interp, "\"", (char *) NULL);
            for (i = 0; i < tree->content.list.n; i++) {
                Tcl_AppendElement(interp, tree->content.list.pntr[i]);
                if (i != (tree->content.list.n - 1))
                    Tcl_AppendResult(interp, "/", (char *) NULL);
            }
            Tcl_AppendResult(interp, "\"", (char *) NULL);
            break;
        }

    case chrChrFnct:
    case numChrFnct:
    case chrNumFnct:
    case numNumFnct:{
            SdbCat                 *next = tree->content.classList;

            SdbPrettyPrint(interp, vars, tree->left);
            Tcl_AppendResult(interp, "?", (char *) NULL);
            while (next) {
                Tcl_AppendResult(interp, "(", NULL);
                if (next->expr) {
                    switch (next->relation) {
                    case opGe:
                        Tcl_AppendResult(interp, ">=", (char *) NULL);
                        break;
                    case opGt:
                        Tcl_AppendResult(interp, ">", (char *) NULL);
                        break;
                    case opLe:
                        Tcl_AppendResult(interp, "<=", (char *) NULL);
                        break;
                    case opLt:
                        Tcl_AppendResult(interp, "<", (char *) NULL);
                        break;
                    case opEq:
                        Tcl_AppendResult(interp, "==", (char *) NULL);
                        break;
                    case opNe:
                        Tcl_AppendResult(interp, "!=", (char *) NULL);
                        break;
                    case opNop:
                        break;
                    default:
                        Tcl_AppendResult(interp, " error! ", (char *) NULL);
                        break;
                    }
                    Tcl_AppendResult(interp, ",", NULL);
                    SdbPrettyPrint(interp, vars, next->expr);
                    Tcl_AppendResult(interp, ",", NULL);
                }
                SdbPrettyPrint(interp, vars, next->code);
                Tcl_AppendResult(interp, ")", NULL);
                if (next->next)
                    Tcl_AppendResult(interp, ":", (char *) NULL);
                next = next->next;
            }
            break;
        }

    case curveFnct:{
            Tcl_AppendResult(interp, "interpolate(", (char *) NULL);
            SdbPrettyPrint(interp, vars, tree->left->left->left);
	    if (tree->left->left->right) {
	      Tcl_AppendResult(interp, "(", NULL);
	      SdbPrettyPrint(interp, vars, tree->left->left->right);
	      Tcl_AppendResult(interp, ")", NULL);
	    }
            Tcl_AppendResult(interp, ",", NULL);
            SdbPrettyPrint(interp, vars, tree->left->right);
            Tcl_AppendResult(interp, ",", NULL);
            SdbPrettyPrint(interp, vars, tree->right->left);
            Tcl_AppendResult(interp, ",", NULL);
            SdbPrettyPrint(interp, vars, tree->right->right);
            Tcl_AppendResult(interp, ")", NULL);
            break;
        }

    case stringFind:
        SdbPrettyPrint(interp, vars, tree->left);
        Tcl_AppendResult(interp, " IN ", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        break;

    case ifNode:
        Tcl_AppendResult(interp, "IF ", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->left);
        Tcl_AppendResult(interp, " Then ", (char *) NULL);
        SdbPrettyPrint(interp, vars, tree->right);
        break;

    default:
        Tcl_AppendResult(interp, "Unknown function!", (char *) NULL);
        break;
    }

    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbInsertLookup  --
 *
 *      Create a new variable that references a lookup table entry
 *
 * Results:
 *      A pointer to the new variable.
 *
 * Side effects:
 *      None
 *
 *------------------------------------------------------------
 */
SdbVariable            *
SdbInsertLookup(dbPtr, string, flags, luHandlePtr, varIndex)
     SdbFile                *dbPtr;
     CONST char             *string;
     int                     flags;
     SdbLookupHandle        *luHandlePtr;
     int                     varIndex;
{

    SdbVariable            *vars;
    SdbNode                *new;

    vars = SdbInsert(dbPtr, string, flags, 0);
    if (!vars)
        return (NULL);

    new = SdbGetNode();
    vars->u.expr[0] = new;
    new->content.luHandlePtr = luHandlePtr;
    if (flags & SDB_CODE) {
        new->type = lookupCode;
    } else {
        new->type = lookupNumber;
    }
    new->right = SdbGetNode();
    (new->right)->type = constCode;
    (new->right)->content.integer = varIndex;

    return (vars);
}


/*
 *------------------------------------------------------------
 *
 * SdbBsave  --
 *
 *      Saves a copy of all allocated variables out ot disk.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      A disk file is created.
 *
 *------------------------------------------------------------
 */
int
SdbBsave(interp, dbPtr, filename)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *filename;
{

    /*
     * Routine to save the database to a diskfile in a binary format.
     * Reloading should be fast because no parsing is required.
     */

    int                     fd,
                            i;
    SdbStack               *top;
    SdbVariable            *varPtr;

    if ((fd = open(filename, O_CREAT | O_WRONLY, 0664)) < 0) {
        Tcl_AppendResult(interp, "could not open file \"", filename,
                "\"", NULL);
        return TCL_ERROR;
    }
    /* write out the signature, and the number of records */
    write(fd, &dbPtr->timePtr, sizeof(dbPtr->timePtr));
    write(fd, &dbPtr->numRec, sizeof(dbPtr->numRec));

    /* The current selection mask... */
    write(fd, dbPtr->select, _bit_size(dbPtr->numRec) * sizeof(_bit_type));

    /* ...and the number of selection stack members, and then
     * the selection stack... */
    for (i = 0, top = dbPtr->stack; top; i++, top = top->next) ;
    write(fd, &i, sizeof(i));
    for (top = dbPtr->stack; top; top = top->next)
        write(fd, top->select, _bit_size(dbPtr->numRec) * sizeof(_bit_type));

    /* ...the current index list, sort list and reverse list pointer... */
    write(fd, dbPtr->list, _bit_size(dbPtr->numRec) * sizeof(*dbPtr->list));
    write(fd, dbPtr->base, _bit_size(dbPtr->numRec) * sizeof(*dbPtr->base));
    write(fd, dbPtr->reverse, _bit_size(dbPtr->numRec) * 
	  sizeof(*dbPtr->reverse));

    /* finally, loop through the variables and save the data values */
    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        i = strlen(varPtr->id);
        write(fd, &i, sizeof(int));

        write(fd, &varPtr->flag, sizeof(varPtr->flag));
        if (varPtr->flag & SDB_ALLOCATED) {
            if (varPtr->flag & SDB_CODE)
                write(fd, varPtr->u.ivalue, dbPtr->numRec * 
		      sizeof(*varPtr->u.ivalue));
            else
                write(fd, varPtr->u.dvalue, dbPtr->numRec * 
		      sizeof(*varPtr->u.dvalue));
        } else if (varPtr->flag & SDB_CONSTANT) {
            if (varPtr->flag & SDB_CODE)
                write(fd, &varPtr->u.code, sizeof(varPtr->u.code));
            else
                write(fd, &varPtr->u.value, sizeof(varPtr->u.value));
        }
    }
    close(fd);
    return TCL_OK;

}


/*
 *------------------------------------------------------------
 *
 * SdbBload  --
 *
 *      Loads a copy of allocated variables from a disk file.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      All old allocated variables are deleted.  A new set is
 *      restored from a disk file.  The number of records may
 *      be different.  There may be different variable names.
 *      The variable dictionary will be refreshed.  
 *
 *------------------------------------------------------------
 */
int
SdbBload(interp, dbPtr, filename)
     Tcl_Interp             *interp;
     SdbFile                *dbPtr;
     char                   *filename;
{

    /*
     * Opposite of 'bsave'.  Loads a database from a binary file.
     */

    int                     fd,
                            i,
                            toggle = 0,
                            idSize,
                            flag;
    SdbStack               *top,
                           *next;
    SdbVariable            *varPtr;
    char                   *id;
    struct timeval          timePtr;
    char                   *fields[2];
    char                   *prev = NULL;

    if ((fd = open(filename, O_RDONLY)) < 0) {
        Tcl_AppendResult(interp, "could not open file \"", filename,
                "\"", NULL);
        return TCL_ERROR;
    }
    /* read the signature, and compare to the existing one.  If it
     * does not match then cancel the load operation.  This is required
     * because the tk_uids will not necessarily match from one execution
     * to another. */
    read(fd, &timePtr, sizeof(timePtr));
    if (bcmp(&dbPtr->timePtr, &timePtr, sizeof(timePtr)) != 0) {
        Tcl_AppendResult(interp, 
			 "binary file does not match the current database",
                NULL);
        close(fd);
        return TCL_ERROR;
    }
    /* read in the number of old records */
    read(fd, &dbPtr->numRec, sizeof(dbPtr->numRec));

    /* replace the current selection mask... */
    if (dbPtr->select)
        ckfree(dbPtr->select);
    dbPtr->select = (_bit_type *) ckalloc(_bit_size(dbPtr->numRec) * 
					  sizeof(_bit_type));
    read(fd, dbPtr->select, _bit_size(dbPtr->numRec) * sizeof(_bit_type));

    /* ...and then the selection stack... (but first clean out the 
     * old selection stack) */
    for (top = dbPtr->stack; top; top = next) {
        next = top->next;
        ckfree(top->select);
        ckfree(top);
    }
    dbPtr->stack = NULL;
    read(fd, &i, sizeof(i));
    while (i) {
        top = dbPtr->stack;
        dbPtr->stack = (SdbStack *) ckalloc(sizeof(SdbStack));

        (dbPtr->stack)->next = top;
        (dbPtr->stack)->select = 
	  (_bit_type *) ckalloc(_bit_size(dbPtr->numRec) *
                sizeof(_bit_type));

        read(fd, &(dbPtr->stack)->select, _bit_size(dbPtr->numRec) *
                sizeof(_bit_type));

        i--;
    }

    /* ...the current index list, sort list and reverse list pointer... */
    if (dbPtr->list)
        ckfree(dbPtr->list);
    dbPtr->list = (int*)ckalloc(dbPtr->numRec * sizeof(*dbPtr->list));
    read(fd, dbPtr->list, _bit_size(dbPtr->numRec) * sizeof(*dbPtr->list));
    if (dbPtr->base)
        ckfree(dbPtr->base);
    dbPtr->base = (int*)ckalloc(dbPtr->numRec * sizeof(*dbPtr->base));
    read(fd, dbPtr->base, _bit_size(dbPtr->numRec) * sizeof(*dbPtr->base));
    if (dbPtr->reverse)
        ckfree(dbPtr->reverse);
    dbPtr->reverse = (int*)ckalloc(dbPtr->numRec * sizeof(*dbPtr->reverse));
    read(fd, dbPtr->reverse, _bit_size(dbPtr->numRec) * 
	 sizeof(*dbPtr->reverse));


    /* finally, loop through the variables and restore the data values */
    while (read(fd, &idSize, sizeof(int)) != sizeof(int)) {
        read(fd, &id, (unsigned) idSize);
        read(fd, &flag, sizeof(int));

        varPtr = SdbFound(dbPtr->varPtr, id);
        if (flag & (SDB_ALLOCATED | SDB_CONSTANT)) {
            /* check to see if the variable being restored already exists.
             * If it does then delete it (whatever it was) and recreate it. */
            if (varPtr)
                SdbDelete(interp, dbPtr, id,1);
            varPtr = SdbInsert(dbPtr, id, flag, dbPtr->numRec);

            /* now read the actual data */
            if (varPtr->flag & SDB_ALLOCATED) {
                if (varPtr->flag & SDB_CODE)
                    read(fd, varPtr->u.ivalue, dbPtr->numRec * 
			 sizeof(*varPtr->u.ivalue));
                else
                    read(fd, varPtr->u.dvalue, dbPtr->numRec * 
			 sizeof(*varPtr->u.dvalue));
            } else if (varPtr->flag & SDB_CONSTANT) {
                if (varPtr->flag & SDB_CODE)
                    read(fd, &varPtr->u.code, sizeof(varPtr->u.code));
                else
                    read(fd, &varPtr->u.value, sizeof(varPtr->u.value));
            }
        }
        /* rearrange the order of variables to what it was previously */
        if (!prev)
            prev = (dbPtr->varPtr)->id;
        fields[0] = prev;
        fields[1] = id;
        /* first time through the current field gets arranged before the
         * very first field in the database.  On subsequent passes the new 
         * fields get arranged after the previous.  */
        SdbArrange(dbPtr, 2, fields, toggle);
        toggle = 1;
        prev = id;

        /* make sure that the output flag is set correctly */
        if (flag & SDB_OUTPUT)
            varPtr->flag |= SDB_OUTPUT;
        else
            varPtr->flag ^= SDB_OUTPUT;
    }
    close(fd);

    SdbReOptimize(dbPtr);
    return TCL_OK;

}

/*
 *------------------------------------------------------------
 *
 * SdbUpdate  --
 *
 *      This is the trace procedure for access to elements of
 *      the database by name.
 *
 * Results:
 *      Returns NULL on sucess.  Returns an error string if
 *      the arguments were incorrect or the action could not be
 *      performed.
 *
 * Side effects:
 *      For read traces call Tcl_SetVar.
 *      For write traces may change the value of a database element.
 *      For unset traces deletes the clientData.
 *
 *------------------------------------------------------------
 */
char                   *
SdbUpdate(clientData, interp, name1, name2, flags)
     ClientData              clientData;
     Tcl_Interp             *interp;
     char                   *name1;
     char                   *name2;
     int                     flags;
{

    SdbVariable            *offset = ((SdbItemData *) clientData)->varPtr;
    SdbFile                *dbPtr = ((SdbItemData *) clientData)->dbPtr;

    SdbVariable            *varPtr;

    char                   *code;
    double                  value;
    int                     row,
                            col;
    char                    tmp_str[256];

    if (offset == NULL) {
        /* offset NULL means that this is the whole array trace.
         * Check the name passed in to see that it is an array reference.
         * If not, just return and let another trace pick this one up.
         * This could be a trace based on a database element defined in
         * the SdbInsert routine, or it could be an invalid element
         * name.
         * 
         * If it is offset -1, then make sure that the array reference
         * is in bounds.
         */
        if (!name2)
            return ((char *) NULL);

        if (sscanf(name2, "%d,%d", &row, &col) != 2)
            return ((char *) NULL);

        /* reject array column and row numbers that are out of range. */
        if ((row < -1 || row >= dbPtr->numRec ||
                        col < -1 || col >= dbPtr->numOutput) &&
                (flags & TCL_TRACE_READS || flags & TCL_TRACE_WRITES)) {
            Tcl_UnsetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
            return ("cell out of range");
        }
        /* 
         * This code moves the database record pointer to track
         * movement of the cursor. This is essential in order to
         * flush the cache correctly.
         */
        if (row >= 0 && row < dbPtr->numRec && row != dbPtr->position) {
            dbPtr->position = row;
            dbPtr->stamp++;
        }
        /* map the real column number into the output array */
        if (col >= 0)
            varPtr = dbPtr->outputPtrPtr[col];
        else
            varPtr = NULL;
    }
    /* This section maps column numbers for requests from the
     * one-dimensional array trace.  In this case the offset
     * contains the real column number to be returned.
     */
    else {
        row = dbPtr->position;
        col = 0;
        varPtr = offset;
    }

    /* handle read traces */
    if (flags & TCL_TRACE_READS) {

        if (col == -1) {
            sprintf(tmp_str, "%d", row);
            Tcl_SetVar2(interp, name1, name2,
                    (char *) tmp_str, flags & TCL_GLOBAL_ONLY);
        } else if (col < -1) {
            /* this column does not exist.  Don't worry, this condition
             * is caught earlier */
        } else if (row == -1) {
            Tcl_SetVar2(interp, name1, name2,
                    varPtr->id, flags & TCL_GLOBAL_ONLY);
        } else if (varPtr == NULL) {
            /* bad show!  how did this happen */
            return ("array variable not defined in this scope");
        } else if (varPtr->flag & SDB_CODE) {
            sprintf(tmp_str, varPtr->format,
                    SdbGetCode(dbPtr, varPtr, dbPtr->base[row]));
            Tcl_SetVar2(interp, name1, name2, tmp_str,
                    flags & TCL_GLOBAL_ONLY);
        } else {
            sprintf(tmp_str, varPtr->format,
                    SdbGetNumber(dbPtr, varPtr, dbPtr->base[row]));
            Tcl_SetVar2(interp, name1, name2, tmp_str,
                    flags & TCL_GLOBAL_ONLY);
        }
    } else if (flags & TCL_TRACE_WRITES) {

        dbPtr->stamp++;
        code = Tcl_GetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);

        /* make sure the pseudo-variables are not being set */
        if (row < 0 || col < 0) {
            Tcl_UnsetVar2(interp, name1, name2, flags & TCL_GLOBAL_ONLY);
            return ("cell out of range");
        } else if (varPtr == NULL) {
            /* bad show!  how did this happen */
            return ("array variable not defined in this scope");
        } else if (varPtr->flag & SDB_CODE) {
            if (varPtr->flag & (SDB_CONSTANT | SDB_ALLOCATED | SDB_LOOKUP))
                SdbPutCode(dbPtr, varPtr, code, dbPtr->base[row]);
            else {
                sprintf(tmp_str, varPtr->format,
                        SdbGetCode(dbPtr, varPtr, dbPtr->base[row]));
                Tcl_SetVar2(interp, name1, name2, tmp_str,
                        flags & TCL_GLOBAL_ONLY);
                return ("can't reset read only variable");
            }
        } else {
            if (sscanf(code, "%lf", &value) != 1)
                value = 0.0;
            if (varPtr->flag & (SDB_CONSTANT | SDB_ALLOCATED | SDB_LOOKUP))
                SdbPutNumber(dbPtr, varPtr, value, dbPtr->base[row]);
            else {
                sprintf(tmp_str, varPtr->format,
                        SdbGetNumber(dbPtr, varPtr, dbPtr->base[row]));
                Tcl_SetVar2(interp, name1, name2, tmp_str,
                        flags & TCL_GLOBAL_ONLY);
                return ("can't reset read only variable");
            }
        }
        /* 
	 * If this cell is a constant, then it might have been optimized
	 * into expressions that use it.  In that case it is essential
	 * to cascade through the up links to update all other cells
         * dependent on this cell.  Don't need to do this if the variable
	 * is mutable.  In that case it will always be referenced to the
	 * source in expressions. 
	 */
	if (!(varPtr->flag & SDB_MUTABLE) &&
	    (varPtr->flag & SDB_CONSTANT)) {
          SdbFreshenUp(dbPtr, varPtr);
         }

    } else if (flags & TCL_TRACE_UNSETS) {
        /* do not delete the trace info if this is an array
         * reference being deleted.  It would be a problem if
         * unset x(1,1) deleted the main trace!  Instead, the
         * database closing routines will take care to delete
         * the trace info (after unsetting the variable).  */
        if (offset != 0)
            ckfree(clientData);
    }
    return ((char *) NULL);
}


/*
 *------------------------------------------------------------
 *
 * SdbRebuildOutput  --
 *
 *      Build an array of variable pointers for quick access by
 *      column.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Allocates a new array.
 *
 *------------------------------------------------------------
 */
void
SdbRebuildOutput(dbPtr)
     SdbFile                *dbPtr;
{

    SdbVariable            *varPtr;
    int                     i;

    dbPtr->numOutput = 0;
    if (dbPtr->outputPtrPtr)
        ckfree(dbPtr->outputPtrPtr);
    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->flag & SDB_OUTPUT) {
            dbPtr->numOutput++;
        }
    }

    i = 0;
    dbPtr->outputPtrPtr =
            (SdbVariable **) ckalloc(dbPtr->numOutput * sizeof(SdbVariable *));

    for (varPtr = dbPtr->varPtr; varPtr; varPtr = varPtr->next) {
        if (varPtr->flag & SDB_OUTPUT) {
            dbPtr->outputPtrPtr[i++] = varPtr;
        }
    }
}


/*
 *------------------------------------------------------------
 *
 * SdbArrange  --
 *
 *      Change the order of columns in the data dictionary.
 *
 * Results:
 *      A standard TCL error code.
 *
 * Side effects:
 *      Rebuilds the output array. 
 *
 *------------------------------------------------------------
 */
int
SdbArrange(dbPtr, argc, argv, flag)
     SdbFile                *dbPtr;
     int                     argc;
     char                  **argv;
     int                     flag;
{
    SdbVariable            *destination,
                           *source,
                          **prev;
    int                     i;

    if (!(destination = SdbFound(dbPtr->varPtr, argv[0]))) {
        Tcl_AppendResult(dbPtr->interp, "variable \"", argv[0], "\" does not exists",
                NULL);
        return TCL_ERROR;
    }
    for (i = 1; i < argc; i++) {
        if (strcmp(argv[0], argv[i]) == 0)
            continue;
        if (!(source = SdbFound(dbPtr->varPtr, argv[i]))) {
            Tcl_AppendResult(dbPtr->interp, "variable \"", argv[i], "\" does not exists",
                    NULL);
	    SdbRebuildOutput(dbPtr);
            return TCL_ERROR;
        }
        /* unlink the source field */
        for (prev = &dbPtr->varPtr;
                *prev != source;
                prev = &((*prev)->next)) ;
        *prev = source->next;

        if (flag) {
            /* move the source *after* the destination */
            source->next = destination->next;
            destination->next = source;
            destination = source;
        } else {
            /* locate the field before the destination */
            for (prev = &dbPtr->varPtr;
                    *prev != destination;
                    prev = &((*prev)->next)) ;
            *prev = source;
            source->next = destination;
        }
    }
    SdbRebuildOutput(dbPtr);
    return TCL_OK;
}


/*
 *------------------------------------------------------------
 *
 * SdbInsert  --
 *
 *      Insert an entry into the data dictionary
 *
 * Results:
 *      A pointer to the new entry, or NULL if there is an error.
 *
 * Side effects:
 *      Rebuilds the output array. 
 *
 *------------------------------------------------------------
 */
SdbVariable            *
SdbInsert(dbPtr, string, flags, estimate)
     SdbFile                *dbPtr;
     CONST char             *string;
     int                     flags;
     int                     estimate;
{
    SdbVariable            *varPtr,
                          **prev;
    int                     new;
    SdbItemData            *id;

/* Tcl_HashEntry *entryPtr; */

    if ((varPtr = SdbFound(dbPtr->varPtr, string))) {
        Tcl_AppendResult(dbPtr->interp, "variable \"", string, 
			 "\" already exists",
                NULL);
        return NULL;
    }
    /* find the last variable in the linked list */
    for (prev = &dbPtr->varPtr;
            *prev;
            prev = &((*prev)->next)) ;

    varPtr = (SdbVariable *) ckalloc(sizeof(SdbVariable));

    if (!varPtr) {
        fprintf(stderr, "could not allocate memory\n");
        exit(1);
    }
    *prev = varPtr;

    varPtr->id = GetIndices(dbPtr->htbl, (char *) string, &new);
    varPtr->flag = flags;
    varPtr->u.integer = 0;
    varPtr->next = NULL;
    varPtr->up = NULL;
    varPtr->down = NULL;
    varPtr->stamp = 0;

    if (flags & SDB_NUMBER)
        varPtr->format = strdup("%-.15g");
    else
        varPtr->format = strdup("%s");


    if (flags & SDB_ALLOCATED) {
        if (flags & SDB_CODE) {
            int                     j;
            char                   *missing = 
	      GetIndices(dbPtr->htbl, ".", &new);

            /*  (*varPtr)[i].u.i.htbl = MALLOC(Tcl_HashTable);
             * Tcl_InitHashTable((*varPtr)[i].u.i.htbl,TCL_STRING_KEYS); */
            if (!(varPtr->u.ivalue = 
		  (char **) ckalloc(estimate * sizeof(char *)))) {
                fprintf(stderr, "could not allocate memory\n");
                exit(1);
            }
            for (j = 0; j < estimate; j++)
                varPtr->u.ivalue[j] = missing;
        } else {
            int                     j;
            if (!(varPtr->u.dvalue = 
		  (double *) ckalloc(estimate * sizeof(double)))) {
                fprintf(stderr, "could not allocate memory\n");
                exit(1);
            }
            for (j = 0; j < estimate; j++)
                varPtr->u.dvalue[j] = 0.0;
        }
    }
    Tcl_SetVar2(dbPtr->interp, dbPtr->dbname, (char *) string, "",
            TCL_GLOBAL_ONLY);
    id = (SdbItemData *) ckalloc(sizeof(SdbItemData));

    id->varPtr = varPtr;
    id->dbPtr = dbPtr;
    Tcl_TraceVar2(dbPtr->interp, dbPtr->dbname, (char *) string,
            TCL_TRACE_WRITES | TCL_TRACE_READS | TCL_TRACE_UNSETS |
            TCL_GLOBAL_ONLY,
            SdbUpdate, (ClientData) id);

    SdbRebuildOutput(dbPtr);

    return varPtr;

}
